package sun.nio.cs;

import java.nio.charset.CodingErrorAction;
import java.nio.CharBuffer;
import java.nio.charset.CoderResult;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.Buffer;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CharsetDecoder;

class CESU_8 extends Unicode
{
    public CESU_8() {
        super("CESU-8", StandardCharsets.aliases_CESU_8);
    }
    
    @Override
    public String historicalName() {
        return "CESU8";
    }
    
    @Override
    public CharsetDecoder newDecoder() {
        return new CESU_8.Decoder(this);
    }
    
    @Override
    public CharsetEncoder newEncoder() {
        return new CESU_8.Encoder(this);
    }
    
    private static final void updatePositions(Buffer buffer, int n, Buffer buffer2, int n2) {
        buffer.position(n - buffer.arrayOffset());
        buffer2.position(n2 - buffer2.arrayOffset());
    }
    
    static void access$200(Buffer buffer, int n, Buffer buffer2, int n2) {
        updatePositions(buffer, n, buffer2, n2);
    }
    
    private static class Decoder extends CharsetDecoder implements ArrayDecoder
    {
        private Decoder(Charset cs) {
            super(cs, 1.0f, 1.0f);
        }
        
        private static boolean isNotContinuation(int n) {
            return (n & 0xC0) != 0x80;
        }
        
        private static boolean isMalformed3(int n, int n2, int n3) {
            return (n == -32 && (n2 & 0xE0) == 0x80) || (n2 & 0xC0) != 0x80 || (n3 & 0xC0) != 0x80;
        }
        
        private static boolean isMalformed3_2(int n, int n2) {
            return (n == -32 && (n2 & 0xE0) == 0x80) || (n2 & 0xC0) != 0x80;
        }
        
        private static boolean isMalformed4(int n, int n2, int n3) {
            return (n & 0xC0) != 0x80 || (n2 & 0xC0) != 0x80 || (n3 & 0xC0) != 0x80;
        }
        
        private static boolean isMalformed4_2(int n, int n2) {
            return (n == 240 && n2 == 144) || (n2 & 0xC0) != 0x80;
        }
        
        private static boolean isMalformed4_3(int n) {
            return (n & 0xC0) != 0x80;
        }
        
        private static CoderResult malformedN(ByteBuffer byteBuffer, int n) {
            switch (n) {
                case 1:
                case 2: {
                    return CoderResult.malformedForLength(1);
                }
                case 3: {
                    int value = byteBuffer.get();
                    int value2 = byteBuffer.get();
                    return CoderResult.malformedForLength(((value == -32 && (value2 & 0xE0) == 0x80) || isNotContinuation(value2)) ? 1 : 2);
                }
                case 4: {
                    int n2 = byteBuffer.get() & 0xFF;
                    int n3 = byteBuffer.get() & 0xFF;
                    if (n2 > 244 || (n2 == 240 && (n3 < 144 || n3 > 191)) || (n2 == 244 && (n3 & 0xF0) != 0x80) || isNotContinuation(n3)) {
                        return CoderResult.malformedForLength(1);
                    }
                    if (isNotContinuation((int)byteBuffer.get())) {
                        return CoderResult.malformedForLength(2);
                    }
                    return CoderResult.malformedForLength(3);
                }
                default: {
                    assert false;
                    return null;
                }
            }
        }
        
        private static CoderResult malformed(ByteBuffer byteBuffer, int n, CharBuffer charBuffer, int n2, int n3) {
            byteBuffer.position(n - byteBuffer.arrayOffset());
            CoderResult malformedN = malformedN(byteBuffer, n3);
            CESU_8.access$200(byteBuffer, n, charBuffer, n2);
            return malformedN;
        }
        
        private static CoderResult malformed(ByteBuffer byteBuffer, int n, int n2) {
            byteBuffer.position(n);
            CoderResult malformedN = malformedN(byteBuffer, n2);
            byteBuffer.position(n);
            return malformedN;
        }
        
        private static CoderResult malformedForLength(ByteBuffer byteBuffer, int n, CharBuffer charBuffer, int n2, int length) {
            CESU_8.access$200(byteBuffer, n, charBuffer, n2);
            return CoderResult.malformedForLength(length);
        }
        
        private static CoderResult malformedForLength(ByteBuffer byteBuffer, int n, int length) {
            byteBuffer.position(n);
            return CoderResult.malformedForLength(length);
        }
        
        private static CoderResult xflow(Buffer buffer, int n, int n2, Buffer buffer2, int n3, int n4) {
            CESU_8.access$200(buffer, n, buffer2, n3);
            return (n4 == 0 || n2 - n < n4) ? CoderResult.UNDERFLOW : CoderResult.OVERFLOW;
        }
        
        private static CoderResult xflow(Buffer buffer, int newPosition, int n) {
            buffer.position(newPosition);
            return (n == 0 || buffer.remaining() < n) ? CoderResult.UNDERFLOW : CoderResult.OVERFLOW;
        }
        
        private CoderResult decodeArrayLoop(ByteBuffer byteBuffer, CharBuffer charBuffer) {
            byte[] array;
            int i;
            int n;
            char[] array2;
            int n2;
            int n3;
            for (array = byteBuffer.array(), i = byteBuffer.arrayOffset() + byteBuffer.position(), n = byteBuffer.arrayOffset() + byteBuffer.limit(), array2 = charBuffer.array(), n2 = charBuffer.arrayOffset() + charBuffer.position(), n3 = charBuffer.arrayOffset() + charBuffer.limit(); n2 < n2 + Math.min(n - i, n3 - n2) && array[i] >= 0; array2[n2++] = (char)array[i++]) {}
            while (i < n) {
                int n4 = array[i];
                if (n4 >= 0) {
                    if (n2 >= n3) {
                        return xflow(byteBuffer, i, n, charBuffer, n2, 1);
                    }
                    array2[n2++] = (char)n4;
                    ++i;
                }
                else if (n4 >> 5 == -2 && (n4 & 0x1E) != 0x0) {
                    if (n - i < 2 || n2 >= n3) {
                        return xflow(byteBuffer, i, n, charBuffer, n2, 2);
                    }
                    int n8 = array[i + 1];
                    if (isNotContinuation(n8)) {
                        return malformedForLength(byteBuffer, i, charBuffer, n2, 1);
                    }
                    array2[n2++] = (char)(n4 << 6 ^ n8 ^ 0xF80);
                    i += 2;
                }
                else if (n4 >> 4 == -2) {
                    int n5 = n - i;
                    if (n5 < 3 || n2 >= n3) {
                        if (n5 > 1 && isMalformed3_2(n4, (int)array[i + 1])) {
                            return malformedForLength(byteBuffer, i, charBuffer, n2, 1);
                        }
                        return xflow(byteBuffer, i, n, charBuffer, n2, 3);
                    }
                    else {
                        int n6 = array[i + 1];
                        int n7 = array[i + 2];
                        if (isMalformed3(n4, n6, n7)) {
                            return malformed(byteBuffer, i, charBuffer, n2, 3);
                        }
                        array2[n2++] = (char)(n4 << 12 ^ n6 << 6 ^ (n7 ^ 0xFFFE1F80));
                        i += 3;
                    }
                }
                else {
                    return malformed(byteBuffer, i, charBuffer, n2, 1);
                }
            }
            return xflow(byteBuffer, i, n, charBuffer, n2, 0);
        }
        
        private CoderResult decodeBufferLoop(ByteBuffer byteBuffer, CharBuffer charBuffer) {
            int i = byteBuffer.position();
            int limit = byteBuffer.limit();
            while (i < limit) {
                int value = byteBuffer.get();
                if (value >= 0) {
                    if (charBuffer.remaining() < 1) {
                        return xflow(byteBuffer, i, 1);
                    }
                    charBuffer.put((char)value);
                    ++i;
                }
                else if (value >> 5 == -2 && (value & 0x1E) != 0x0) {
                    if (limit - i < 2 || charBuffer.remaining() < 1) {
                        return xflow(byteBuffer, i, 2);
                    }
                    int value4 = byteBuffer.get();
                    if (isNotContinuation(value4)) {
                        return malformedForLength(byteBuffer, i, 1);
                    }
                    charBuffer.put((char)(value << 6 ^ value4 ^ 0xF80));
                    i += 2;
                }
                else if (value >> 4 == -2) {
                    int n = limit - i;
                    if (n < 3 || charBuffer.remaining() < 1) {
                        if (n > 1 && isMalformed3_2(value, (int)byteBuffer.get())) {
                            return malformedForLength(byteBuffer, i, 1);
                        }
                        return xflow(byteBuffer, i, 3);
                    }
                    else {
                        int value2 = byteBuffer.get();
                        int value3 = byteBuffer.get();
                        if (isMalformed3(value, value2, value3)) {
                            return malformed(byteBuffer, i, 3);
                        }
                        charBuffer.put((char)(value << 12 ^ value2 << 6 ^ (value3 ^ 0xFFFE1F80)));
                        i += 3;
                    }
                }
                else {
                    return malformed(byteBuffer, i, 1);
                }
            }
            return xflow(byteBuffer, i, 0);
        }
        
        @Override
        protected CoderResult decodeLoop(ByteBuffer byteBuffer, CharBuffer charBuffer) {
            if (byteBuffer.hasArray() && charBuffer.hasArray()) {
                return this.decodeArrayLoop(byteBuffer, charBuffer);
            }
            return this.decodeBufferLoop(byteBuffer, charBuffer);
        }
        
        private static ByteBuffer getByteBuffer(ByteBuffer wrap, byte[] array, int n) {
            if (wrap == null) {
                wrap = ByteBuffer.wrap(array);
            }
            wrap.position(n);
            return wrap;
        }
        
        @Override
        public int decode(byte[] array, int i, int a, char[] array2) {
            int n = i + a;
            int n2 = 0;
            int min = Math.min(a, array2.length);
            ByteBuffer byteBuffer = null;
            while (n2 < min && array[i] >= 0) {
                array2[n2++] = (char)array[i++];
            }
            while (i < n) {
                int n3 = array[i++];
                if (n3 >= 0) {
                    array2[n2++] = (char)n3;
                }
                else if (n3 >> 5 == -2 && (n3 & 0x1E) != 0x0) {
                    if (i < n) {
                        int n6 = array[i++];
                        if (isNotContinuation(n6)) {
                            if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                                return -1;
                            }
                            array2[n2++] = this.replacement().charAt(0);
                            --i;
                        }
                        else {
                            array2[n2++] = (char)(n3 << 6 ^ n6 ^ 0xF80);
                        }
                    }
                    else {
                        if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                            return -1;
                        }
                        array2[n2++] = this.replacement().charAt(0);
                        return n2;
                    }
                }
                else if (n3 >> 4 == -2) {
                    if (i + 1 < n) {
                        int n4 = array[i++];
                        int n5 = array[i++];
                        if (isMalformed3(n3, n4, n5)) {
                            if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                                return -1;
                            }
                            array2[n2++] = this.replacement().charAt(0);
                            i -= 3;
                            byteBuffer = getByteBuffer(byteBuffer, array, i);
                            i += malformedN(byteBuffer, 3).length();
                        }
                        else {
                            array2[n2++] = (char)(n3 << 12 ^ n4 << 6 ^ (n5 ^ 0xFFFE1F80));
                        }
                    }
                    else {
                        if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                            return -1;
                        }
                        if (i < n && isMalformed3_2(n3, (int)array[i])) {
                            array2[n2++] = this.replacement().charAt(0);
                        }
                        else {
                            array2[n2++] = this.replacement().charAt(0);
                            return n2;
                        }
                    }
                }
                else {
                    if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                        return -1;
                    }
                    array2[n2++] = this.replacement().charAt(0);
                }
            }
            return n2;
        }
    }
    
    private static class Encoder extends CharsetEncoder implements ArrayEncoder
    {
        private Surrogate.Parser sgp;
        private char[] c2;
        
        private Encoder(Charset cs) {
            super(cs, 1.1f, 3.0f);
        }
        
        @Override
        public boolean canEncode(char ch) {
            return !Character.isSurrogate(ch);
        }
        
        @Override
        public boolean isLegalReplacement(byte[] repl) {
            return (repl.length == 1 && repl[0] >= 0) || super.isLegalReplacement(repl);
        }
        
        private static CoderResult overflow(CharBuffer charBuffer, int n, ByteBuffer byteBuffer, int n2) {
            CESU_8.access$200(charBuffer, n, byteBuffer, n2);
            return CoderResult.OVERFLOW;
        }
        
        private static CoderResult overflow(CharBuffer charBuffer, int n) {
            charBuffer.position(n);
            return CoderResult.OVERFLOW;
        }
        
        private static void to3Bytes(byte[] array, int n, char c) {
            array[n] = (byte)(0xE0 | c >> 12);
            array[n + 1] = (byte)(0x80 | (c >> 6 & 0x3F));
            array[n + 2] = (byte)(0x80 | (c & '?'));
        }
        
        private static void to3Bytes(ByteBuffer byteBuffer, char c) {
            byteBuffer.put((byte)(0xE0 | c >> 12));
            byteBuffer.put((byte)(0x80 | (c >> 6 & 0x3F)));
            byteBuffer.put((byte)(0x80 | (c & '?')));
        }
        
        private CoderResult encodeArrayLoop(CharBuffer charBuffer, ByteBuffer byteBuffer) {
            char[] array;
            int i;
            int n;
            byte[] array2;
            int n2;
            int n3;
            for (array = charBuffer.array(), i = charBuffer.arrayOffset() + charBuffer.position(), n = charBuffer.arrayOffset() + charBuffer.limit(), array2 = byteBuffer.array(), n2 = byteBuffer.arrayOffset() + byteBuffer.position(), n3 = byteBuffer.arrayOffset() + byteBuffer.limit(); n2 < n2 + Math.min(n - i, n3 - n2) && array[i] < '\u0080'; array2[n2++] = (byte)array[i++]) {}
            while (i < n) {
                int ch = array[i];
                if (ch < '\u0080') {
                    if (n2 >= n3) {
                        return overflow(charBuffer, i, byteBuffer, n2);
                    }
                    array2[n2++] = (byte)ch;
                }
                else if (ch < 'ࠀ') {
                    if (n3 - n2 < 2) {
                        return overflow(charBuffer, i, byteBuffer, n2);
                    }
                    array2[n2++] = (byte)(0xC0 | ch >> 6);
                    array2[n2++] = (byte)(0x80 | (ch & '?'));
                }
                else if (Character.isSurrogate((char)ch)) {
                    if (this.sgp == null) {
                        this.sgp = new Surrogate.Parser();
                    }
                    int parse = this.sgp.parse((char)ch, array, i, n);
                    if (parse < 0) {
                        CESU_8.access$200(charBuffer, i, byteBuffer, n2);
                        return this.sgp.error();
                    }
                    if (n3 - n2 < 6) {
                        return overflow(charBuffer, i, byteBuffer, n2);
                    }
                    to3Bytes(array2, n2, Character.highSurrogate(parse));
                    n2 += 3;
                    to3Bytes(array2, n2, Character.lowSurrogate(parse));
                    n2 += 3;
                    ++i;
                }
                else {
                    if (n3 - n2 < 3) {
                        return overflow(charBuffer, i, byteBuffer, n2);
                    }
                    to3Bytes(array2, n2, (char)ch);
                    n2 += 3;
                }
                ++i;
            }
            CESU_8.access$200(charBuffer, i, byteBuffer, n2);
            return CoderResult.UNDERFLOW;
        }
        
        private CoderResult encodeBufferLoop(CharBuffer charBuffer, ByteBuffer byteBuffer) {
            int position;
            for (position = charBuffer.position(); charBuffer.hasRemaining(); ++position) {
                int value = charBuffer.get();
                if (value < '\u0080') {
                    if (!byteBuffer.hasRemaining()) {
                        return overflow(charBuffer, position);
                    }
                    byteBuffer.put((byte)value);
                }
                else if (value < 'ࠀ') {
                    if (byteBuffer.remaining() < 2) {
                        return overflow(charBuffer, position);
                    }
                    byteBuffer.put((byte)(0xC0 | value >> 6));
                    byteBuffer.put((byte)(0x80 | (value & '?')));
                }
                else if (Character.isSurrogate((char)value)) {
                    if (this.sgp == null) {
                        this.sgp = new Surrogate.Parser();
                    }
                    int parse = this.sgp.parse((char)value, charBuffer);
                    if (parse < 0) {
                        charBuffer.position(position);
                        return this.sgp.error();
                    }
                    if (byteBuffer.remaining() < 6) {
                        return overflow(charBuffer, position);
                    }
                    to3Bytes(byteBuffer, Character.highSurrogate(parse));
                    to3Bytes(byteBuffer, Character.lowSurrogate(parse));
                    ++position;
                }
                else {
                    if (byteBuffer.remaining() < 3) {
                        return overflow(charBuffer, position);
                    }
                    to3Bytes(byteBuffer, (char)value);
                }
            }
            charBuffer.position(position);
            return CoderResult.UNDERFLOW;
        }
        
        @Override
        protected final CoderResult encodeLoop(CharBuffer charBuffer, ByteBuffer byteBuffer) {
            if (charBuffer.hasArray() && byteBuffer.hasArray()) {
                return this.encodeArrayLoop(charBuffer, byteBuffer);
            }
            return this.encodeBufferLoop(charBuffer, byteBuffer);
        }
        
        @Override
        public int encode(char[] array, int i, int a, byte[] array2) {
            int n = i + a;
            int n2;
            for (n2 = 0; n2 < n2 + Math.min(a, array2.length) && array[i] < '\u0080'; array2[n2++] = (byte)array[i++]) {}
            while (i < n) {
                int ch = array[i++];
                if (ch < '\u0080') {
                    array2[n2++] = (byte)ch;
                }
                else if (ch < 'ࠀ') {
                    array2[n2++] = (byte)(0xC0 | ch >> 6);
                    array2[n2++] = (byte)(0x80 | (ch & '?'));
                }
                else if (Character.isSurrogate((char)ch)) {
                    if (this.sgp == null) {
                        this.sgp = new Surrogate.Parser();
                    }
                    int parse = this.sgp.parse((char)ch, array, i - 1, n);
                    if (parse < 0) {
                        if (this.malformedInputAction() != CodingErrorAction.REPLACE) {
                            return -1;
                        }
                        array2[n2++] = this.replacement()[0];
                    }
                    else {
                        to3Bytes(array2, n2, Character.highSurrogate(parse));
                        n2 += 3;
                        to3Bytes(array2, n2, Character.lowSurrogate(parse));
                        n2 += 3;
                        ++i;
                    }
                }
                else {
                    to3Bytes(array2, n2, (char)ch);
                    n2 += 3;
                }
            }
            return n2;
        }
    }
}
